from ngsolve import Mesh
from ngstents import TentSlab
from netgen.geom2d import unit_square


def test_tent_nlayers():
    mesh = Mesh(unit_square.GenerateMesh(maxh=.3))
    dt = 5
    c = 1
    global_ct = 0.999
    method = "vol"
    heapsize = 5 * 1000 * 1000
    tentslab = TentSlab(mesh, method, heapsize)
    tentslab.SetMaxWavespeed(c)
    tentslab.PitchTents(dt, global_ct=global_ct)
    tents = [tentslab.GetTent(i) for i in range(tentslab.GetNTents())]
    layers = set([t.level for t in tents])
    assert len(layers) == tentslab.GetNLayers(), "Incorrect number of layers"



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html



# -- Project information -----------------------------------------------------

project = 'ngstents'
copyright = '2023, Jay Gopalakrishnan'
author = 'Jay Gopalakrishnan'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc","sphinx.ext.mathjax","sphinx.ext.todo","sphinx.ext.githubpages",
              "IPython.sphinxext.ipython_console_highlighting", "IPython.sphinxext.ipython_directive",
              "jupyter_sphinx",
              "nbsphinx",
              "myst_parser",
              "sphinxemoji.sphinxemoji",
              "myst_nb.sphinx_ext",
              ]

# Run notebook configuration

# The template used when exporting from nbconvert
#   full  - Outputs the full HTML document [Default]
#   basic - Outputs a single div (with no additional resources)
run_notebook_export_template = 'basic'  # Default: 'full'

# Display the source links to the generated evaluated files
run_notebook_display_source_links = False  # Default: True

# Whether or not to evaluate the notebooks prior to embedding them
evaluate_notebooks = True  # Default: True

# START nbsphinx stuff
#increase timeout for cell execution, since some files take long to execute
nbsphinx_timeout = 100000

# If True, the build process is continued even if an exception occurs:
nbsphinx_allow_errors = False

# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = r"""
.. raw:: html

    <style>
        .p-Widget {
            height: 400px;
        }
        .dg.main {
            margin-left: 0px;
        }
        div.p-Widget div div div div.dg ul li {
            list-style: none;
            margin-left: 0px;
        }
        div.p-Widget div div div div.dg ul li div.dg {
            margin-bottom: 0px;
        }

        .MathJax { font-size: 0.9em !important; }
    </style>

.. only:: html
    .. role:: raw-html(raw)
        :format: html
"""

nbsphinx_widgets_path = ""
# END nbsphinx stuff

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'paper', 'env', 'jupyter_execute']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

html_context = {
  'display_github': True,
  'github_user': 'jayggg',
  'github_repo': 'ngstents',
}



from netgen.geom2d import SplineGeometry
from ngsolve import (Mesh, Draw, Redraw, CoefficientFunction, sqrt, sin, cos,
                     x, y, TaskManager, L2, GridFunction, Integrate, InnerProduct)
from ngsolve.internal import visoptions
from ngstents import TentSlab
from ngstents.conslaw import Wave
from math import pi
import time

geom = SplineGeometry()
geom.AddRectangle(p1=(0, 0), p2=(pi, pi), bc="reflect")
mesh = Mesh(geom.GenerateMesh(maxh=0.25))

# setting the problem
period = sqrt(2)*pi
n_periods = 1
t_end = period*n_periods
dt = t_end/10
heapsize = 10*1000*1000
wavespeed = 1

# using causality constant
local_ctau = True
global_ctau = 2/3
ts = TentSlab(mesh, method="edge", heapsize=heapsize)
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

# setting the approximation space order
order = 2
V = L2(mesh, order=order, dim=mesh.dim+1)
u = GridFunction(V,"u")
wave = Wave(u, ts, reflect=mesh.Boundaries("reflect"))
wave.SetTentSolver("SAT", stages=order+1, substeps=4*order)

mu0 = CoefficientFunction(cos(x)*cos(y))
q0 = CoefficientFunction((0, 0))
cf = CoefficientFunction((q0, mu0))
wave.SetInitial(cf)

Draw(u)
visoptions.scalfunction = "u:3"

t = 0
cnt = 0

redraw = 1
# input("start")
t1 = time.time()
with TaskManager():
    while t < t_end - dt/2:
        wave.Propagate()
        t += dt
        cnt += 1
        if cnt % redraw == 0:
            print("{:.3f}".format(t))
            Redraw(True)
print("total time = {}".format(time.time()-t1))

# exact solution
exsol = CoefficientFunction((sin(x)*cos(y)*sin(sqrt(2)*t_end)/sqrt(2),
                             cos(x)*sin(y)*sin(sqrt(2)*t_end)/sqrt(2),
                             cos(x)*cos(y)*cos(sqrt(2)*t_end)))
Draw(exsol, mesh, "exact")
l2error = sqrt(Integrate(InnerProduct(u-exsol, u-exsol), mesh, order=3*order))
print("l2error = ", l2error)



from netgen.geom2d import SplineGeometry
from ngsolve import (Mesh, Draw, Redraw, CoefficientFunction, IfPos, sqrt, sin,
                     cos, exp, x, y, TaskManager, L2, GridFunction, Integrate,
                     InnerProduct)
from ngsolve.internal import visoptions
from ngstents import TentSlab
from ngstents.conslaw import Wave
import time

geom = SplineGeometry()
geom.AddRectangle(p1=(0, 0), p2=(2, 2), bc="square")
mesh = geom.GenerateMesh(maxh=0.25)
mesh = Mesh(mesh)

# setting the problem
t_end = 2
dt = t_end / 10
wavespeed = 1

# using causality constant
local_ctau = True
global_ctau = 2 / 3
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

# setting the approximation space order
order = 2
V = L2(mesh, order=order, dim=mesh.dim + 1)
u = GridFunction(V, "u")
wave = Wave(u, ts)
wave.SetTentSolver("SAT", stages=order + 1, substeps=4 * order)
# wave.SetTentSolver("SARK", substeps=4*order)


def f(s):
    return (exp(-10 * (s - 1) * (s - 1)) - exp(-10)) / (1 - exp(-10)) * IfPos(
        s, IfPos(2 - s, 1, 0), 0)


k = CoefficientFunction((cos(1), sin(1)))
pos = CoefficientFunction((x, y))


def uex(time):
    return CoefficientFunction((k, 1)) * f(time - InnerProduct(k, pos))


mu0 = CoefficientFunction(0)
q0 = CoefficientFunction((0, 0))
cf = CoefficientFunction((q0, mu0))
wave.SetInitial(cf)

t = wave.time  # advancing front
wave.SetBoundaryCF(mesh.BoundaryCF({"square": uex(t)}))

Draw(u)
visoptions.scalfunction = "u:3"

t = 0
cnt = 0

redraw = 1
# input("start")
t1 = time.time()
with TaskManager():
    while t < t_end - dt / 2:
        wave.Propagate()
        t += dt
        cnt += 1
        if cnt % redraw == 0:
            print("{:.3f}".format(t))
            Redraw(True)
print("total time = {}".format(time.time() - t1))

# exact solution
uexact = uex(t_end)
Draw(uexact, mesh, "u_exact")
l2error = sqrt(
    Integrate(InnerProduct(u - uexact, u - uexact), mesh, order=3 * order))
print("l2error = ", l2error)



from ._pyconslaw import ConservationLaw


class Burgers(ConservationLaw):
    def __init__(self, gfu, tentslab, **kwargs):
        """
        INPUTS:

        gridfunction : GridFunction
        tentslab     : TentSlab
        outflow      : Optional[Region]
        inflow       : Optional[Region]
        reflect      : Optional[Region]
        """
        ConservationLaw.__init__(self, gfu, tentslab, "burgers", **kwargs)


class Euler(ConservationLaw):
    def __init__(self, gfu, tentslab, **kwargs):
        """
        INPUTS:

        gridfunction : GridFunction
        tentslab     : TentSlab
        outflow      : Optional[Region]
        inflow       : Optional[Region]
        reflect      : Optional[Region]
        """
        ConservationLaw.__init__(self, gfu, tentslab, "euler", **kwargs)


class Wave(ConservationLaw):
    def __init__(self, gfu, tentslab, **kwargs):
        """
        INPUTS:

        gridfunction : GridFunction
        tentslab     : TentSlab
        outflow      : Optional[Region]
        inflow       : Optional[Region]
        reflect      : Optional[Region]
        transparent  : Optional[Region]
        """
        ConservationLaw.__init__(self, gfu, tentslab, "wave", **kwargs)


class Advection(ConservationLaw):
    def __init__(self, gfu, tentslab, **kwargs):
        """
        INPUTS:

        gridfunction : GridFunction
        tentslab     : TentSlab
        outflow      : Optional[Region]
        inflow       : Optional[Region]
        reflect      : Optional[Region]
        transparent  : Optional[Region]
        """
        ConservationLaw.__init__(self, gfu, tentslab, "advection", **kwargs)


class Maxwell(ConservationLaw):
    def __init__(self, gfu, tentslab, **kwargs):
        """
        INPUTS:

        gridfunction : GridFunction
        tentslab     : TentSlab
        outflow      : Optional[Region]
        inflow       : Optional[Region]
        reflect      : Optional[Region]
        """
        ConservationLaw.__init__(self, gfu, tentslab, "maxwell", **kwargs)



import sys

sys.path.append("../demo/wave")
sys.path.append("../demo/advection")
sys.path.append("../demo/symbolic")

def test_wave2d():
    ''' 
    reference values generated with NGSolve-6.2.2102-17-g02efd4be3
    and ngstents commit 55862b9
    
    standing wave example:
    * mesh size h = 0.25, spatial order = 2
    * structure-aware Taylor time stepping with 3 stages and 8 substeps within each tent
    * reference for spatial L2 error: 3.60383e-04
    '''
    import wave2d
    assert wave2d.l2error <= 4e-4

def test_wave2d_timdepbc():
    '''
    plane wave example:
    * mesh size h = 0.25, spatial order = 2
    * structure-aware Taylor time stepping with 3 stages and 8 substeps within each tent
    * reference for spatial L2 error: 1.11393e-02
    '''
    import wave2d_timedepbc
    assert wave2d_timedepbc.l2error <= 2e-2
    

def test_advection2d():
    ''' 
    reference values generated with NGSolve-6.2.2102-17-g02efd4be3
    and ngstents commit 55862b9
    
    standing wave example:
    * mesh size h = 0.15, spatial order = 4
    * structure-aware Taylor time stepping with 5 stages and 8 substeps within each tent
    * reference for spatial L2 error: 2.79097e-03
    '''
    import advection2d
    assert advection2d.l2error <= 3e-3

def test_symbolic_wave():
    ''' 
    reference values generated with NGSolve-6.2.2102-17-g02efd4be3
    and ngstents commit 55862b9
    
    standing wave example:
    * mesh size h = 0.25, spatial order = 2
    * structure-aware Runge-Kutta time stepping with 8 substeps within each tent
    * reference for spatial L2 error: 3.78149e-04
    '''
    import symbolic_wave
    assert symbolic_wave.l2error <= 4e-4

if __name__ == "__main__":
    functions = [test_wave2d, test_wave2d_timdepbc,
                 test_advection2d, test_symbolic_wave]
    passed = []
    print("Test wave equation:")
    for func in functions:
        try:
            func()
            passed.append(True)
        except AssertionError:
            passed.append(False)
    print("Function name \t\t pass/fail")
    for i in range(len(functions)):
        if passed[i]:
            print(functions[i].__name__,": \t\t pass")
        else:
            print(functions[i].__name__,": \t\t fail")



"""
Burgers equation in 2-D
"""
from netgen.geom2d import unit_square
from ngsolve import Mesh, CoefficientFunction, x, y, exp, Draw, Redraw
from ngsolve import L2, GridFunction
from ngsolve import TaskManager, SetNumThreads
from ngstents import TentSlab
from ngstents.conslaw import Burgers

# options
step_sol = False
sol_vec = False
vtk_tents = False
ngs_gui = False

maxh = 0.05
mesh = Mesh(unit_square.GenerateMesh(maxh=maxh))

heapsize = 10*1000*1000
dt = 0.05
c = 6
ts = TentSlab(mesh, heapsize=heapsize)
ts.SetMaxWavespeed(c)
ts.PitchTents(dt)
print("max slope", ts.MaxSlope())

order = 4
V = L2(mesh, order=order)
u = GridFunction(V,"u")
burg = Burgers(u, ts, outflow=mesh.Boundaries("left|bottom|right|top"))
burg.SetTentSolver("SARK",substeps=order*order)

cf = CoefficientFunction(exp(-50*((x-0.3)*(x-0.3)+(y-0.3)*(y-0.3))))
burg.SetInitial(cf)

Draw(u)

tend = 10*dt
t = 0
cnt = 0
if ngs_gui:
    Draw(burg)

if vtk_tents:
    burg.VTKTents('temp')

input('start')
with TaskManager():
    while t < tend - dt/2:
        burg.Propagate()
        t += dt
        cnt += 1
        if cnt % 1 == 0:
            print("{:.3f}".format(t))
            Redraw(True)
            if step_sol:
                input('step')

if sol_vec:
    print(sol.vec)



"""
OBSOLETE / NEEDS REVISION ...
"""
from ngsolve import Mesh, Draw, Redraw
from ngsolve import CoefficientFunction, cos, x, InnerProduct, Id
from ngsolve import L2, GridFunction, TaskManager
from ngsolve import specialcf as scf
from ngsolve.internal import visoptions, viewoptions
from ngstents import TentSlab
from ngstents.meshes import Make1DMesh
from ngstents.conslaw import ConservationLaw
from math import pi

N = 128
mesh = Mesh(Make1DMesh([[0, 2], [2, 4]], [N, N], bcname=["left", "right"]))
'''
solve wave equation: d_tt(phi) - div(alpha grad(phi)) = 0
define first order system

d_t(g(u)) - div f(u) = 0

for u = [q,mu]^t with q = -alpha grad(phi) and mu = d_t(phi)
g(u) = [[alpha^-1 0],[0, 1]] * u
f(u) = [[ I*mu ],
        [ q^t  ]]
isentropic material: alpha = I*c^2 for wave speed c
'''

# wave speeds in domains
ws = CoefficientFunction([1, 1 / 2])

tend = 7
dt = 0.1
local_ctau = True
global_ctau = 1 / 2
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(ws)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

order = 2
V = L2(mesh, order=order, dim=mesh.dim + 1)
gfu = GridFunction(V, name="u")
# vector-valued coefficient function
n = scf.normal(mesh.dim)


def Flux(u):
    """
    Compute the flux for given TrialFunction u,
    a matrix-valued CoefficientFunction of size (h,w) where
    h represents the number of equations and
    w should correspond to the dimension of the mesh/space
    """
    return CoefficientFunction((Id(mesh.dim) * u[mesh.dim], u[0:mesh.dim]),
                               dims=(V.dim, mesh.dim))


def NumFlux(um, up):
    """
    Compute the numeric flux
    Inputs
    um: trace of u for current element on facet
    up: trace of u for other element on facet
    """
    # upwind flux
    flux = 0.5 * (Flux(um) + Flux(up)) * n
    flux_vec = 0.5 * (um[0:mesh.dim] - up[0:mesh.dim]) * n * n
    flux_scal = 0.5 * (um[mesh.dim] - up[mesh.dim])
    return flux + CoefficientFunction((flux_vec, flux_scal))


def InverseMap(y):
    """
    solves "y = u - (f(u),gradphi)" for u
    """
    norm_sqr = InnerProduct(ts.gradphi, ts.gradphi)
    mu = (y[mesh.dim] + ws**2 * InnerProduct(y[0:mesh.dim], ts.gradphi)) / (
        1 - ws**2 * norm_sqr)
    q = ws**2 * (y[0:mesh.dim] + mu * ts.gradphi)
    return CoefficientFunction((q, mu))


def BndNumFlux_transparent(um):
    """
    defines numerical flux on boundary elements using the normal vector n
    um: trace of u for current element on facet
    """
    return CoefficientFunction(
        (0.5 * (um[mesh.dim] + 1 / ws * um[0:mesh.dim] * n) * n,
         0.5 * (um[mesh.dim] + (2 - 1 / ws) * um[0:mesh.dim] * n)))


cl = ConservationLaw(gfu,
                     ts,
                     flux=Flux,
                     numflux=NumFlux,
                     inversemap=InverseMap)
cl.SetTentSolver("SAT", stages=order + 1, substeps=4 * order)
# cl.SetTentSolver("SARK", substeps=4*order)

# set inital data
cl.SetInitial(CoefficientFunction((0, 0)))


# define plane wave for boundary condition
def f(s):
    return (CoefficientFunction(1) - cos(4 * pi * (s))) * IfPos(
        s, IfPos(1 / 2 - s, 1, 0), 0)


def planewave(t):
    return CoefficientFunction((ws**2, ws)) * f(t - x)


t = cl.time  # advancing front
cl.SetBoundaryCF(
    mesh.BoundaryCF({
        "left": NumFlux(cl.u_minus, planewave(t)),
        "right": BndNumFlux_transparent(cl.u_minus)
    }))

Draw(gfu)
visoptions.scalfunction = "u:2"
viewoptions.drawedges = 1

redraw = 1
t = 0
cnt = 0
input("start")
import time

t1 = time.time()
with TaskManager():
    while t < tend - dt / 2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt % redraw == 0:
            print("{:5f}".format(t))
        Redraw(True)
print("total time = ", time.time() - t1)



import ngsolve as ng
from netgen.geom2d import EdgeInfo as EI, Solid2d, CSG2d
from ngstents import TentSlab
from ngstents.conslaw import Wave
from ngsolve import L2, GridFunction
from ngsolve.internal import visoptions
from ngsolve import exp, CF, y


geo = CSG2d()

horn = Solid2d(
    [(1, 0.55),
     EI((1, 1), bc='out'),      # right curved boundary (with control point)
     (0, 1),
     EI((-1,  1), bc='out'),    # left curved bdry
     (-1, 0.55),
     EI(bc='cone'),             # conical walls
     (-0.03, 0),
     EI(maxh=0.02, bc='pipe'),  # feed pipe
     (-0.03, -0.5),
     EI(maxh=0.02, bc='in'),
     (+0.03, -0.5),
     EI(maxh=0.02, bc='pipe'),
     (+0.03, 0),
     EI(bc='cone')
     ], mat='air')

geo.Add(horn)
m = geo.GenerateMesh(maxh=0.15)
mesh = ng.Mesh(m)
mesh.Curve(4)
# ng.Draw(mesh)


ts = TentSlab(mesh, method="edge", heapsize=10*1000*1000)
wavespeed = 1
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=0.1, local_ct=True, global_ct=2/3)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())


order = 4
V = L2(mesh, order=order, dim=mesh.dim+1)
u = GridFunction(V, "u")
wave = Wave(u, ts,
            inflow=mesh.Boundaries('in'),
            transparent=mesh.Boundaries('out'),
            reflect=mesh.Boundaries('pipe|cone'))

wave.SetTentSolver("SAT", stages=order+1, substeps=4*order)


def f(s):
    d = 500
    fs = exp(-s**2 * d)
    dfs = -2 * d * s * fs
    return fs, dfs


phi, dphi = f(y+0.2)
q0 = CF((0, -dphi))
mu0 = -dphi
wave.SetInitial(CF((q0, mu0)))
ng.Draw(u)
visoptions.scalfunction = 'u:3'
visoptions.subdivisions = '4'
visoptions.autoscale = False
visoptions.mmaxval = 8
visoptions.mminval = -8

t = 0
cnt = 0
dt = 0.1
redraw = 1

with ng.TaskManager():
    while t < 0.7:
        wave.Propagate()
        t += dt
        cnt += 1

        # input("continue?")

        if cnt % redraw == 0:
            print("{:.3f}".format(t))
            ng.Redraw(True)



from netgen.geom2d import SplineGeometry
from ngsolve import Mesh, Draw, Redraw
from ngsolve import (CoefficientFunction, IfPos, cos, x, y, InnerProduct,
                     OuterProduct, Id)
from ngsolve import L2, GridFunction, TaskManager
from ngsolve import specialcf as scf
from ngsolve.internal import visoptions
from ngstents import TentSlab
from ngstents.conslaw import ConservationLaw
from math import pi
import time

# define geometry
a = 0.25
geom = SplineGeometry()
pnts = [(-1, 0), (-a, 0), (-a, a), (0, a), (a, a), (a, 0), (1, 0), (1, 1),
        (-1, 1)]
segs = [(0, 1), (1, 5), (5, 6), (6, 7), (7, 8), (8, 0), (1, 2, 3), (3, 4, 5)]
bcs = ["bottom", "bottom", "bottom", "right", "top", "left", "cyl", "cyl"]
dominout = [None, (2, 0), None, None, None, None, (1, 2), (1, 2)]
pind = [geom.AddPoint(*pnt) for pnt in pnts]
for i, seg in enumerate(segs):
    if (dominout[i]):
        leftdomain = dominout[i][0]
        rightdomain = dominout[i][1]
    else:
        leftdomain = 1
        rightdomain = 0
    if (len(seg) == 3):
        geom.Append(['spline3', *seg],
                    leftdomain=leftdomain,
                    rightdomain=rightdomain)
    else:
        geom.Append(['line', *seg],
                    bc=bcs[i],
                    leftdomain=leftdomain,
                    rightdomain=rightdomain)
maxh = 0.1
geom.SetDomainMaxH(2, maxh / 2)
ngmesh = geom.GenerateMesh(maxh=maxh)
mesh = Mesh(ngmesh)
mesh.Curve(3)
'''
solve wave equation: d_tt(phi) - div(alpha grad(phi)) = 0
define first order system

d_t(g(u)) - div f(u) = 0

for u = [q,mu]^t with q = -alpha grad(phi) and mu = d_t(phi)
g(u) = [[alpha^-1 0],[0, 1]] * u
f(u) = [[ I*mu ],
        [ q^t  ]]
isentropic material: alpha = I*c^2 for wave speed c
'''

tend = 2.5
dt = 0.025
# define wave speed
# c = 1   for |x| > a (domain1)
# c = 1/2 for |x| < a (domain2)
wavespeed = CoefficientFunction([1, 1 / 2])

# using causality constant
local_ctau = True
global_ctau = 1 / 4
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

# vector-valued coefficient function
n = scf.normal(mesh.dim)


def Flux(u):
    """
    Compute the flux f(u) for given TrialFunction u,
    a matrix-valued CoefficientFunction of size (h,w) where
    h represents the number of equations and
    w should correspond to the dimension of the mesh/space
    """
    return CoefficientFunction((Id(mesh.dim) * u[mesh.dim], u[0:mesh.dim]),
                               dims=(V.dim, mesh.dim))


def NumFlux(um, up):
    """
    Compute the numeric flux
    Inputs
    um: trace of u for current element on facet
    up: trace of u for other element on facet
    """
    # upwind flux
    flux = 0.5 * (Flux(um) + Flux(up)) * n
    flux_vec = 0.5 * (um[0:mesh.dim] - up[0:mesh.dim]) * n * n
    flux_scal = 0.5 * (um[mesh.dim] - up[mesh.dim])
    return flux + CoefficientFunction((flux_vec, flux_scal))


def InverseMap(y):
    """
    solves "y = g(u) - (f(u),gradphi)" for u
    g(u) = [[1/c^2 0],[0, 1]] * u
    """
    norm_sqr = InnerProduct(ts.gradphi, ts.gradphi)
    ip = InnerProduct(y[0:mesh.dim], ts.gradphi)
    mu = (y[mesh.dim] + wavespeed**2 * ip) / (1 - wavespeed**2 * norm_sqr)
    q = wavespeed**2 * (y[0:mesh.dim] + mu * ts.gradphi)
    return CoefficientFunction((q, mu))


def ReflectBnd(u):
    """
    reflects values of u at the boundary using the normal vector n
    """
    q = u[0:mesh.dim]
    mu = u[mesh.dim]
    return CoefficientFunction((q - 2 * OuterProduct(n, n) * q, mu))


def BndNumFlux(um):
    """
    defines numerical flux on boundary elements using the normal vector n
    um: trace of u for current element on facet
    """
    # set (q,n) = 0
    return CoefficientFunction((um[mesh.dim] * n, 0))


# define conservation law
order = 2
V = L2(mesh, order=order, dim=mesh.dim + 1)
gfu = GridFunction(V, name="u")
cl = ConservationLaw(gfu,
                     ts,
                     flux=Flux,
                     numflux=NumFlux,
                     inversemap=InverseMap)
cl.SetTentSolver("SARK", substeps=4 * order)
# cl.SetTentSolver("SAT", stages=order+1, substeps=4*order)

# set inital data
mu0 = CoefficientFunction(0)
q0 = CoefficientFunction((0, 0))
cl.SetInitial(CoefficientFunction((q0, mu0)))


# define plane wave for boundary condition
def f(s):
    return (CoefficientFunction(1) - cos(4 * pi * (s - 1))) * IfPos(
        s - 1, IfPos(3 / 2 - s, 1, 0), 0)


k = CoefficientFunction((1, 0))
pos = CoefficientFunction((x, y))


def uex(time):
    return CoefficientFunction((k, 1)) * f(time - InnerProduct(k, pos))


cl.SetBoundaryCF(
    mesh.BoundaryCF({
        "left": NumFlux(cl.u_minus, uex(cl.time)),
        "bottom|right|top": BndNumFlux(cl.u_minus)
    }))

Draw(gfu)
visoptions.scalfunction = "u:3"

t = 0
cnt = 0
redraw = 1

input("press enter to start")
t1 = time.time()
with TaskManager():
    while t < tend - dt / 2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt % redraw == 0:
            print("{:5f}".format(t))
            Redraw(True)
print("total time = ", time.time() - t1)



from netgen.geom2d import SplineGeometry
from ngsolve import Mesh, Draw, Redraw
from ngsolve import Integrate
from ngsolve import CoefficientFunction, exp, sqrt, sin, cos, x, y, z
from ngsolve import L2, GridFunction
from ngsolve import TaskManager
from ngsolve.internal import visoptions
from ngstents import TentSlab
from ngstents.conslaw import Advection
from math import pi
import time

geom = SplineGeometry()
geom.AddCircle(c=(0,0),r=1,bc="circle")
mesh = geom.GenerateMesh(maxh=0.15)
mesh = Mesh(mesh)

# setting the problem
tend = 10
dt = tend/10
wavespeed = 2

# using causality constant
local_ctau = True
global_ctau = 1
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

order = 4
V = L2(mesh, order=order)
gfu = GridFunction(V,"u")
cl = Advection(gfu, ts, inflow=mesh.Boundaries("circle"))
cl.SetVectorField( CoefficientFunction((y,-x)) )
cl.SetTentSolver("SAT",stages=order+1, substeps=2*order)

u0 = CoefficientFunction( exp(-100* ((x-0.5)*(x-0.5)+y*y)))
cl.SetInitial(u0)

Draw(gfu)
visoptions.autoscale = 0

t = 0
cnt = 0
redraw = 1

import time
# input("press enter to start")
t1 = time.time()
with TaskManager():
    while t < tend-dt/2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt%redraw == 0:
            print("{:5f}".format(t))
        Redraw(True)
print("total time = ",time.time()-t1)

# exact solution
exsol = CoefficientFunction(exp(-100*(
    (x-0.5*cos(tend))*(x-0.5*cos(tend))
    +(y+0.5*sin(tend))*(y+0.5*sin(tend))
)))

Draw(exsol,mesh,"exact")
error = gfu - exsol
Draw(error,mesh,"error")
visoptions.scalfunction = "u:"

l2error = sqrt(Integrate(error*error,mesh,order=3*order))
print("l2error = ", l2error)
print("ndof = ", cl.space.ndof)



import time
from math import pi
from ngstents.conslaw import Wave
from ngstents import TentSlab
from ngsolve.internal import visoptions, viewoptions
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from ngsolve import (Mesh, Draw, Redraw, CoefficientFunction, sqrt, sin, cos,
                     x, y, z, TaskManager, L2, GridFunction, Integrate, InnerProduct)


geom = CSGeometry()
brick = OrthoBrick(Pnt(0, 0, 0), Pnt(pi, pi, pi)).bc("reflect")
geom.Add(brick)
mesh = geom.GenerateMesh(maxh=0.5)
mesh = Mesh(mesh)

# setting the problem
period = sqrt(3)*pi
n_periods = 1
t_end = period*n_periods
dt = t_end/12
heapsize = 10*1000*1000
wavespeed = 1

# using causality constant
local_ctau = True
global_ctau = 1/2
ts = TentSlab(mesh, method="edge", heapsize=heapsize)
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

# setting the approximation space order
order = 2
V = L2(mesh, order=order, dim=mesh.dim+1)
u = GridFunction(V,"u")
wave = Wave(u, ts, reflect=mesh.Boundaries("reflect"))
wave.SetTentSolver("SAT", stages=order+1, substeps=4*order)

mu0 = CoefficientFunction(cos(x)*cos(y)*cos(z))
q0 = CoefficientFunction((0, 0, 0))
cf = CoefficientFunction((q0, mu0))
wave.SetInitial(cf)

Draw(u)
visoptions.scalfunction = "u:4"
viewoptions.clipping.enable = 1
visoptions.clipsolution = 'scal'
viewoptions.clipping.dist = 0.5

t = 0
cnt = 0

redraw = 1
input('start')
t1 = time.time()
with TaskManager():
    print("starting...")
    while t < t_end - dt/2:
        wave.Propagate()
        t += dt
        cnt += 1
        if cnt % redraw == 0:
            print("{:.3f}".format(t))
            Redraw(True)
print("total time = {}".format(time.time()-t1))
# exact solution
exsol = CoefficientFunction((sin(x)*cos(y)*cos(z)*sin(sqrt(3)*t_end)/sqrt(3),
                             cos(x)*sin(y)*cos(z)*sin(sqrt(3)*t_end)/sqrt(3),
                             cos(x)*cos(y)*sin(z)*sin(sqrt(3)*t_end)/sqrt(3),
                             cos(x)*cos(y)*cos(z)*cos(sqrt(3)*t_end)))
Draw(exsol, mesh, 'exact')
l2error = sqrt(Integrate(InnerProduct(u-exsol, u-exsol), mesh, order=3*order))
print("l2error = ", l2error)



from netgen.geom2d import SplineGeometry
from ngsolve import (Mesh, L2, GridFunction, CoefficientFunction,
                     exp, x, y, TaskManager)
from ngstents import TentSlab
from ngstents.conslaw import Burgers


def test_burgers2D():
    order = 2
    dt = 0.025
    c = 16
    tend = dt
    method = "edge"

    geom = SplineGeometry()
    geom.AddRectangle((0, 0), (1, 1), bc=1)
    mesh = Mesh(geom.GenerateMesh(maxh=0.2))
    ts = TentSlab(mesh, method)
    ts.SetMaxWavespeed(c)
    success = ts.PitchTents(dt)
    try:
        assert success is True
    except AssertionError as e:
        msg = "Slab could not be pitched"
        e.args += ("Failed to pitch slab", msg)
        raise
    cf = CoefficientFunction(exp(-50*((x-0.3)*(x-0.3)+(y-0.3)*(y-0.3))))

    V = L2(mesh, order=order)
    u = GridFunction(V, "u")
    burg = Burgers(u, ts)
    burg.SetTentSolver("SARK", substeps=order*order)
    burg.SetInitial(cf)

    t = 0
    with TaskManager():
        while t < tend - dt/2:
            burg.Propagate()
            t += dt
